@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2015-2017 Cray Inc. All rights reserved.
- * Copyright (c) 2015-2017 Los Alamos National Security, LLC.
+ * Copyright (c) 2015-2018 Los Alamos National Security, LLC.
  *                         All rights reserved.
  *
  * This software is available to you under a choice of one of two
@@ -173,6 +173,14 @@ typedef ssize_t (*trecvmsg_func_t)(struct fid_ep *ep,
  */
 int  _gnix_ep_int_tx_pool_grow(struct gnix_fid_ep *ep);
 
+/**
+ * Internal function for initializing tx buffer pool
+ *
+ * @param[in] ep	pointer to a EP
+ */
+int _gnix_ep_int_tx_pool_init(struct gnix_fid_ep *ep);
+
+
 /*
  * inline functions
  */
@@ -228,7 +236,7 @@ _gnix_fr_alloc(struct gnix_fid_ep *ep)
 	struct gnix_fab_req *fr = NULL;
 	int ret = _gnix_fl_alloc(&de, &ep->fr_freelist);
 
-	while (unlikely(ret == -FI_EAGAIN))
+	while (OFI_UNLIKELY(ret == -FI_EAGAIN))
 		ret = _gnix_fl_alloc(&de, &ep->fr_freelist);
 
 	if (ret == FI_SUCCESS) {
@@ -252,7 +260,7 @@ _gnix_fr_alloc_w_cb(struct gnix_fid_ep *ep, void (*cb)(void *))
 	struct gnix_fab_req *fr = NULL;
 	int ret = _gnix_fl_alloc(&de, &ep->fr_freelist);
 
-	while (unlikely(ret == -FI_EAGAIN))
+	while (OFI_UNLIKELY(ret == -FI_EAGAIN))
 		ret = _gnix_fl_alloc(&de, &ep->fr_freelist);
 
 	if (ret == FI_SUCCESS) {
@@ -276,6 +284,8 @@ _gnix_fr_free(struct gnix_fid_ep *ep, struct gnix_fab_req *fr)
 {
 	assert(fr->gnix_ep == ep);
 
+	assert((fr->flags & FI_LOCAL_MR) == 0);
+
 	if (fr->int_tx_buf_e != NULL) {
 		_gnix_ep_release_int_tx_buf(ep, fr->int_tx_buf_e);
 		fr->int_tx_buf_e = NULL;
@@ -348,11 +358,18 @@ int _gnix_ep_alloc(struct fid_domain *domain, struct fi_info *info,
 int _gnix_ep_init_vc(struct gnix_fid_ep *ep_priv);
 
 /**
- * Internal function for enabling an ep
+ * Internal function for enabling ep tx resources
  *
  * @param[in] ep_priv	 pointer to a previously allocated EP
  */
-int _gnix_ep_enable(struct gnix_fid_ep *ep_priv);
+int _gnix_ep_tx_enable(struct gnix_fid_ep *ep_priv);
+
+/**
+ * Internal function for enabling ep rx resources
+ *
+ * @param[in] ep_priv	 pointer to a previously allocated EP
+ */
+int _gnix_ep_rx_enable(struct gnix_fid_ep *ep_priv);
 
 /*******************************************************************************
  * API Functions
