@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2015-2017 Cray Inc. All rights reserved.
- * Copyright (c) 2015-2017 Los Alamos National Security, LLC.
+ * Copyright (c) 2015-2018 Los Alamos National Security, LLC.
  *                         All rights reserved.
  *
  * This software is available to you under a choice of one of two
@@ -74,7 +74,7 @@ int _gnix_ep_int_tx_pool_grow(struct gnix_fid_ep *ep)
 	struct gnix_fid_mem_desc *md = NULL;
 	struct gnix_int_tx_buf *tx_buf_list;
 	struct gnix_int_tx_ptrs *tx_ptrs;
-
+	
 	assert(ep);
 
 	if (ep->int_tx_pool.nbufs >= GNIX_INT_TX_POOL_COUNT) {
@@ -101,11 +101,12 @@ int _gnix_ep_int_tx_pool_grow(struct gnix_fid_ep *ep)
 		goto tx_ptrs_err;
 	}
 
-	ret = gnix_mr_reg(&ep->domain->domain_fid.fid, tx_bufs,
+	ret = _gnix_mr_reg(&ep->domain->domain_fid.fid, tx_bufs,
 			GNIX_INT_TX_BUF_SZ * GNIX_INT_TX_POOL_SIZE,
-			FI_READ | FI_WRITE, 0, 0, 0, &auto_mr, NULL);
+			FI_READ | FI_WRITE, 0, 0, 0,
+			&auto_mr, NULL, ep->auth_key, GNIX_PROV_REG);
 
-	if (unlikely(ret != FI_SUCCESS)) {
+	if (OFI_UNLIKELY(ret != FI_SUCCESS)) {
 		GNIX_WARN(FI_LOG_EP_DATA, "gnix_mr_req returned: %s\n",
 			   fi_strerror(-ret));
 		goto reg_err;
@@ -235,21 +236,8 @@ static void __fr_freelist_destroy(struct gnix_fid_ep *ep)
 	_gnix_fl_destroy(&ep->fr_freelist);
 }
 
-int _gnix_ep_enable(struct gnix_fid_ep *ep)
+int _gnix_ep_rx_enable(struct gnix_fid_ep *ep)
 {
-	/*
-	 * set enable flags and add nic to poll lists
-	 */
-
-	if (ep->send_cq) {
-		_gnix_cq_poll_obj_add(ep->send_cq, ep->nic,
-				      _gnix_nic_progress);
-		if (ep->cm_nic) /* No CM NIC for MSG EPs */
-			_gnix_cq_poll_obj_add(ep->send_cq, ep->cm_nic,
-					      _gnix_cm_nic_progress);
-		ep->tx_enabled = true;
-	}
-
 	if (ep->recv_cq) {
 		_gnix_cq_poll_obj_add(ep->recv_cq, ep->nic,
 				      _gnix_nic_progress);
@@ -259,38 +247,6 @@ int _gnix_ep_enable(struct gnix_fid_ep *ep)
 		ep->rx_enabled = true;
 	}
 
-	if (ep->send_cntr) {
-		_gnix_cntr_poll_obj_add(ep->send_cntr, ep->nic,
-					_gnix_nic_progress);
-		if (ep->cm_nic) /* No CM NIC for MSG EPs */
-			_gnix_cntr_poll_obj_add(ep->send_cntr, ep->cm_nic,
-						_gnix_cm_nic_progress);
-	}
-
-	if (ep->recv_cntr) {
-		_gnix_cntr_poll_obj_add(ep->recv_cntr, ep->nic,
-					_gnix_nic_progress);
-		if (ep->cm_nic) /* No CM NIC for MSG EPs */
-			_gnix_cntr_poll_obj_add(ep->recv_cntr, ep->cm_nic,
-						_gnix_cm_nic_progress);
-	}
-
-	if (ep->write_cntr) {
-		_gnix_cntr_poll_obj_add(ep->write_cntr, ep->nic,
-					_gnix_nic_progress);
-		if (ep->cm_nic) /* No CM NIC for MSG EPs */
-			_gnix_cntr_poll_obj_add(ep->write_cntr, ep->cm_nic,
-						_gnix_cm_nic_progress);
-	}
-
-	if (ep->read_cntr) {
-		_gnix_cntr_poll_obj_add(ep->read_cntr, ep->nic,
-					_gnix_nic_progress);
-		if (ep->cm_nic) /* No CM NIC for MSG EPs */
-			_gnix_cntr_poll_obj_add(ep->read_cntr, ep->cm_nic,
-						_gnix_cm_nic_progress);
-	}
-
 	if (ep->rwrite_cntr) {
 		_gnix_cntr_poll_obj_add(ep->rwrite_cntr, ep->nic,
 					_gnix_nic_progress);
@@ -310,6 +266,46 @@ int _gnix_ep_enable(struct gnix_fid_ep *ep)
 	return FI_SUCCESS;
 }
 
+int _gnix_ep_tx_enable(struct gnix_fid_ep *ep)
+{
+
+	if (ep->send_cq) {
+		_gnix_cq_poll_obj_add(ep->send_cq, ep->nic,
+				      _gnix_nic_progress);
+		if (ep->cm_nic) /* No CM NIC for MSG EPs */
+			_gnix_cq_poll_obj_add(ep->send_cq, ep->cm_nic,
+					      _gnix_cm_nic_progress);
+		ep->tx_enabled = true;
+	}
+
+	if (ep->send_cntr) {
+		_gnix_cntr_poll_obj_add(ep->send_cntr, ep->nic,
+					_gnix_nic_progress);
+		if (ep->cm_nic) /* No CM NIC for MSG EPs */
+			_gnix_cntr_poll_obj_add(ep->send_cntr, ep->cm_nic,
+						_gnix_cm_nic_progress);
+	}
+
+	if (ep->write_cntr) {
+		_gnix_cntr_poll_obj_add(ep->write_cntr, ep->nic,
+					_gnix_nic_progress);
+		if (ep->cm_nic) /* No CM NIC for MSG EPs */
+			_gnix_cntr_poll_obj_add(ep->write_cntr, ep->cm_nic,
+						_gnix_cm_nic_progress);
+	}
+
+	if (ep->read_cntr) {
+		_gnix_cntr_poll_obj_add(ep->read_cntr, ep->nic,
+					_gnix_nic_progress);
+		if (ep->cm_nic) /* No CM NIC for MSG EPs */
+			_gnix_cntr_poll_obj_add(ep->read_cntr, ep->cm_nic,
+						_gnix_cm_nic_progress);
+	}
+
+	return FI_SUCCESS;
+}
+
+
 /*******************************************************************************
  * Forward declaration for ops structures
  ******************************************************************************/
@@ -1565,10 +1561,18 @@ DIRECT_FN STATIC int gnix_ep_control(fid_t fid, int command, void *arg)
 			}
 		}
 
-		ret = _gnix_ep_enable(ep);
+		ret = _gnix_ep_tx_enable(ep);
 		if (ret != FI_SUCCESS) {
 			GNIX_WARN(FI_LOG_EP_CTRL,
-			     "_gnix_ep_enable call returned %d\n",
+			     "_gnix_ep_tx_enable call returned %d\n",
+				ret);
+			goto err;
+		}
+
+		ret = _gnix_ep_rx_enable(ep);
+		if (ret != FI_SUCCESS) {
+			GNIX_WARN(FI_LOG_EP_CTRL,
+			     "_gnix_ep_rx_enable call returned %d\n",
 				ret);
 			goto err;
 		}
@@ -2306,7 +2310,8 @@ DIRECT_FN int gnix_ep_open(struct fid_domain *domain, struct fi_info *info,
 
 	if (info->ep_attr->auth_key_size) {
 		auth_key = GNIX_GET_AUTH_KEY(info->ep_attr->auth_key,
-				info->ep_attr->auth_key_size);
+				info->ep_attr->auth_key_size,
+				domain_priv->using_vmdh);
 		if (!auth_key)
 			return -FI_EINVAL;
 	} else {
@@ -2482,7 +2487,8 @@ int _gnix_ep_alloc(struct fid_domain *domain, struct fi_info *info,
 
 	if (info->ep_attr->auth_key_size) {
 		auth_key = GNIX_GET_AUTH_KEY(info->ep_attr->auth_key,
-				info->ep_attr->auth_key_size);
+				info->ep_attr->auth_key_size,
+				domain_priv->using_vmdh);
 		if (!auth_key)
 			return -FI_EINVAL;
 	} else {
@@ -2869,6 +2875,11 @@ DIRECT_FN STATIC ssize_t gnix_ep_cancel(fid_t fid, void *context)
 		_gnix_cntr_inc_err(err_cntr);
 	}
 
+	if (req->flags & FI_LOCAL_MR) {
+		fi_close(&req->amo.loc_md->mr_fid.fid);
+		req->flags &= ~FI_LOCAL_MR;
+	}
+
 	_gnix_fr_free(ep, req);
 
 	return ret;
