@@ -45,14 +45,14 @@
 #include <stdio.h>
 #include <assert.h>
 
-#include <fi_util.h>
+#include <ofi_util.h>
 #include <rdma/fabric.h>
 #include <rdma/fi_cm.h>
 #include <rdma/fi_domain.h>
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_rma.h>
 #include <rdma/fi_errno.h>
-#include "prov.h"
+#include "ofi_prov.h"
 
 #include "gnix.h"
 #include "gnix_nic.h"
@@ -63,6 +63,32 @@
 #include "gnix_wait.h"
 #include "gnix_xpmem.h"
 
+/* check if only one bit of a set is enabled, when one is required */
+#define IS_EXCLUSIVE(x) \
+	((x) && !((x) & ((x)-1)))
+
+/* optional basic bits */
+#define GNIX_MR_BASIC_OPT \
+	(FI_MR_LOCAL)
+
+/* optional scalable bits */
+#define GNIX_MR_SCALABLE_OPT \
+	(FI_MR_LOCAL)
+
+/* required basic bits */
+#define GNIX_MR_BASIC_REQ \
+	(FI_MR_VIRT_ADDR | FI_MR_ALLOCATED | FI_MR_PROV_KEY)
+
+/* required scalable bits */
+#define GNIX_MR_SCALABLE_REQ \
+	(FI_MR_MMU_NOTIFY)
+
+#define GNIX_MR_BASIC_BITS \
+	(GNIX_MR_BASIC_OPT | GNIX_MR_BASIC_REQ)
+
+#define GNIX_MR_SCALABLE_BITS \
+	(GNIX_MR_SCALABLE_OPT | GNIX_MR_SCALABLE_REQ)
+
 const char gnix_fab_name[] = "gni";
 const char gnix_dom_name[] = "/sys/class/gni/kgni0";
 const char gnix_prov_name[] = "gni";
@@ -99,17 +125,17 @@ static int __gnix_auth_key_get_val(
 
 #define GNIX_DEFAULT_USER_REGISTRATION_LIMIT 192
 #define GNIX_DEFAULT_PROV_REGISTRATION_LIMIT 64
+#define GNIX_DEFAULT_SHARED_MEMORY_TIMEOUT 30
 
 int gnix_default_user_registration_limit = GNIX_DEFAULT_USER_REGISTRATION_LIMIT;
 int gnix_default_prov_registration_limit = GNIX_DEFAULT_PROV_REGISTRATION_LIMIT;
+uint32_t gnix_wait_shared_memory_timeout = GNIX_DEFAULT_SHARED_MEMORY_TIMEOUT;
 
 /* assume that the user will open additional fabrics later and that
    ptag information will need to be retained for the lifetime of the
    process. If the user sets this value, we can assume that they
    intend to be done with libfabric when the last fabric instance
    closes so that we can free the ptag information. */
-/* TODO: implement this such that the user give us more information
-	about teardown */
 int gnix_dealloc_aki_on_fabric_close = 0;
 
 const struct fi_fabric_attr gnix_fabric_attr = {
@@ -245,7 +271,7 @@ static struct fi_info *_gnix_allocinfo(void)
 
 	gnix_info->domain_attr->name = strdup(gnix_dom_name);
 	gnix_info->domain_attr->cq_data_size = sizeof(uint64_t);
-	gnix_info->domain_attr->mr_mode = FI_MR_BASIC | OFI_MR_BASIC_MAP;
+	gnix_info->domain_attr->mr_mode = FI_MR_BASIC;
 	gnix_info->domain_attr->resource_mgmt = FI_RM_ENABLED;
 	gnix_info->domain_attr->mr_key_size = sizeof(uint64_t);
 	gnix_info->domain_attr->max_ep_tx_ctx = GNIX_SEP_MAX_CNT;
@@ -277,7 +303,7 @@ static struct fi_info *_gnix_allocinfo(void)
 }
 
 static int __gnix_getinfo_resolve_node(const char *node, const char *service,
-				       uint64_t flags, struct fi_info *hints,
+				       uint64_t flags, const struct fi_info *hints,
 				       struct fi_info *info)
 {
 	int ret;
@@ -290,7 +316,7 @@ static int __gnix_getinfo_resolve_node(const char *node, const char *service,
 		is_fi_addr_str = true;
 	}
 
-	if (unlikely(is_fi_addr_str && node && service)) {
+	if (OFI_UNLIKELY(is_fi_addr_str && node && service)) {
 		GNIX_WARN(FI_LOG_FABRIC, "service parameter must be NULL when "
 			"node parameter is not and using FI_ADDR_STR.\n");
 		return -FI_EINVAL;
@@ -418,12 +444,13 @@ err:
 
 static int _gnix_ep_getinfo(enum fi_ep_type ep_type, uint32_t version,
 			    const char *node, const char *service,
-			    uint64_t flags, struct fi_info *hints,
+			    uint64_t flags, const struct fi_info *hints,
 			    struct fi_info **info)
 {
 	uint64_t mode = GNIX_FAB_MODES;
 	struct fi_info *gnix_info = NULL;
 	int ret = -FI_ENODATA;
+	int mr_mode;
 
 	GNIX_TRACE(FI_LOG_FABRIC, "\n");
 
@@ -536,6 +563,8 @@ static int _gnix_ep_getinfo(enum fi_ep_type ep_type, uint32_t version,
 		GNIX_DEBUG(FI_LOG_FABRIC, "Passed fabric name check\n");
 
 		if (hints->domain_attr) {
+			mr_mode = hints->domain_attr->mr_mode;
+
 			if (hints->domain_attr->name &&
 			    strncmp(hints->domain_attr->name, gnix_dom_name,
 				    strlen(gnix_dom_name))) {
@@ -552,19 +581,47 @@ static int _gnix_ep_getinfo(enum fi_ep_type ep_type, uint32_t version,
 				gnix_info->domain_attr->data_progress =
 					hints->domain_attr->data_progress;
 
-			if (ofi_check_mr_mode(version,
+			/* If basic registration isn't being requested,
+			   require FI_MR_MMU_NOTIFY */
+			if (!(hints->domain_attr->mr_mode &
+					(FI_MR_BASIC | FI_MR_ALLOCATED)))
+				gnix_info->domain_attr->mr_mode |= FI_MR_MMU_NOTIFY;
+
+			if (ofi_check_mr_mode(&gnix_prov, version,
 					gnix_info->domain_attr->mr_mode,
-					hints->domain_attr->mr_mode) != FI_SUCCESS) {
-				GNIX_DEBUG(FI_LOG_FABRIC,
-					"failed mr_mode check\n");
+					hints) != FI_SUCCESS) {
+				GNIX_INFO(FI_LOG_DOMAIN,
+					"failed ofi_check_mr_mode, "
+					"ret=%d\n", ret);
 				goto err;
 			}
 
-			if (FI_VERSION_LT(version, FI_VERSION(1, 5)))
-				gnix_info->domain_attr->mr_mode = FI_MR_BASIC;
-			else
-				gnix_info->domain_attr->mr_mode =
-					hints->domain_attr->mr_mode;
+			if (FI_VERSION_LT(version, FI_VERSION(1, 5))) {
+				switch (mr_mode) {
+				case FI_MR_UNSPEC:
+				case FI_MR_BASIC:
+					mr_mode = FI_MR_BASIC;
+					break;
+				default:
+					GNIX_DEBUG(FI_LOG_FABRIC,
+						"unsupported mr_mode selected, "
+						"ret=%d\n", ret);
+					goto err;
+				}
+			} else {
+				/* define the mode we return to the user
+				 * prefer basic until scalable
+				 * has more testing time */
+				if (mr_mode & FI_MR_BASIC)
+					mr_mode = OFI_MR_BASIC_MAP;
+				else if ((mr_mode & GNIX_MR_BASIC_REQ) ==
+							GNIX_MR_BASIC_REQ)
+					mr_mode &= GNIX_MR_BASIC_BITS;
+				else
+					mr_mode &= GNIX_MR_SCALABLE_BITS;
+			}
+
+			gnix_info->domain_attr->mr_mode = mr_mode;
 
 			switch (hints->domain_attr->threading) {
 			case FI_THREAD_COMPLETION:
@@ -586,16 +643,6 @@ static int _gnix_ep_getinfo(enum fi_ep_type ep_type, uint32_t version,
 					hints->domain_attr->caps;
 			}
 
-			ret = ofi_check_domain_attr(&gnix_prov, version,
-						    gnix_info->domain_attr,
-						    hints->domain_attr);
-			if (ret) {
-				GNIX_WARN(FI_LOG_FABRIC,
-						  "GNI failed domain attributes check\n");
-				goto err;
-			}
-
-			GNIX_DEBUG(FI_LOG_FABRIC, "Passed the domain attributes check\n");
 		}
 	}
 
@@ -606,10 +653,12 @@ static int _gnix_ep_getinfo(enum fi_ep_type ep_type, uint32_t version,
 
 	ofi_alter_info(gnix_info, hints, version);
 
+	GNIX_DEBUG(FI_LOG_FABRIC, "Passed the domain attributes check\n");
+
 	/* The provider may silently enable secondary
 	 * capabilities that do not introduce any overhead. */
 	if (hints && hints->caps)
-	    gnix_info->caps = hints->caps | GNIX_EP_SEC_CAPS;
+		gnix_info->caps = hints->caps | GNIX_EP_SEC_CAPS;
 	else
 		gnix_info->caps = GNIX_EP_CAPS_FULL | GNIX_EP_SEC_CAPS;
 
@@ -631,7 +680,7 @@ err:
 }
 
 static int gnix_getinfo(uint32_t version, const char *node, const char *service,
-			uint64_t flags, struct fi_info *hints,
+			uint64_t flags, const struct fi_info *hints,
 			struct fi_info **info)
 {
 	int ret = 0;
@@ -673,7 +722,7 @@ static void gnix_fini(void)
 struct fi_provider gnix_prov = {
 	.name = gnix_prov_name,
 	.version = FI_VERSION(GNI_MAJOR_VERSION, GNI_MINOR_VERSION),
-	.fi_version = FI_VERSION(1, 5),
+	.fi_version = FI_VERSION(1, 6),
 	.getinfo = gnix_getinfo,
 	.fabric = gnix_fabric_open,
 	.cleanup = gnix_fini
@@ -685,6 +734,7 @@ GNI_INI
 	gni_return_t status;
 	gni_version_info_t lib_version;
 	int num_devices;
+	int ret;
 
 	/*
 	 * if no GNI devices available, don't register as provider
@@ -714,7 +764,10 @@ GNI_INI
 	}
 
 	/* Initialize global MR notifier. */
-	_gnix_notifier_init();
+	ret = _gnix_smrn_init();
+	if (ret != FI_SUCCESS)
+		GNIX_FATAL(FI_LOG_FABRIC,
+			"failed to initialize global mr notifier\n");
 
 	/* Initialize global NIC data. */
 	_gnix_nic_init();
@@ -747,6 +800,9 @@ __gnix_fab_ops_get_val(struct fid *fid, fab_ops_val_t t, void *val)
 	case GNI_DEFAULT_PROV_REGISTRATION_LIMIT:
 		*(uint32_t *)val = gnix_default_prov_registration_limit;
 		break;
+	case GNI_WAIT_SHARED_MEMORY_TIMEOUT:
+		*(uint32_t *)val = gnix_wait_shared_memory_timeout;
+		break;
 	default:
 		GNIX_WARN(FI_LOG_FABRIC, ("Invalid fab_ops_val\n"));
 	}
@@ -793,6 +849,10 @@ __gnix_fab_ops_set_val(struct fid *fid, fab_ops_val_t t, void *val)
 		}
 		gnix_default_prov_registration_limit = v;
 		break;
+	case GNI_WAIT_SHARED_MEMORY_TIMEOUT:
+		v = *(uint32_t *) val;
+		gnix_wait_shared_memory_timeout = v;
+		break;
 	default:
 		GNIX_WARN(FI_LOG_FABRIC, ("Invalid fab_ops_val\n"));
 		return -FI_EINVAL;
@@ -903,6 +963,11 @@ static int __gnix_auth_key_set_val(
 		GNIX_WARN(FI_LOG_FABRIC,
 			"GNIX_TOTAL_KEYS_NEEDED is not a definable value.\n");
 		return -FI_EOPNOTSUPP;
+	case GNIX_USER_KEY_MAX_PER_RANK:
+		GNIX_WARN(FI_LOG_FABRIC,
+			"GNIX_USER_KEY_MAX_PER_RANK is not a definable "
+			"value.\n");
+		return -FI_EOPNOTSUPP;
 	default:
 		GNIX_WARN(FI_LOG_FABRIC, ("Invalid fab_ops_val\n"));
 		return -FI_EINVAL;
@@ -918,6 +983,8 @@ static int __gnix_auth_key_get_val(
 	void *val)
 {
 	struct gnix_auth_key *info;
+	uint32_t pes_on_node;
+	int ret;
 
 	if (!val)
 		return -FI_EINVAL;
@@ -941,7 +1008,20 @@ static int __gnix_auth_key_get_val(
 			info->attr.prov_key_limit) :
 			(gnix_default_user_registration_limit +
 			 gnix_default_prov_registration_limit));
-	break;
+		break;
+	case GNIX_USER_KEY_MAX_PER_RANK:
+		ret = _gnix_pes_on_node(&pes_on_node);
+		if (ret) {
+			GNIX_WARN(FI_LOG_FABRIC,
+				"failed to get pes on node count\n");
+			return -FI_EINVAL;
+		}
+
+		*(int *)val = ((info) ?
+			info->attr.user_key_limit :
+			gnix_default_user_registration_limit) /
+			pes_on_node;
+		break;
 	default:
 		GNIX_WARN(FI_LOG_FABRIC, ("Invalid fab_ops_val\n"));
 		return -FI_EINVAL;
