@@ -142,18 +142,19 @@ ssize_t sock_ep_recvmsg(struct fid_ep *ep, const struct fi_msg *msg,
 static ssize_t sock_ep_recv(struct fid_ep *ep, void *buf, size_t len,
 				void *desc, fi_addr_t src_addr, void *context)
 {
-	struct fi_msg msg;
-	struct iovec msg_iov;
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = buf;
-	msg_iov.iov_len = len;
+	struct iovec msg_iov = {
+		.iov_base = buf,
+		.iov_len = len,
+	};
+	struct fi_msg msg = {
+		.msg_iov = &msg_iov,
+		.desc = &desc,
+		.iov_count = 1,
+		.addr = src_addr,
+		.context = context,
+		.data = 0,
+	};
 
-	msg.msg_iov = &msg_iov;
-	msg.desc = &desc;
-	msg.iov_count = 1;
-	msg.addr = src_addr;
-	msg.context = context;
-	msg.data = 0;
 	return sock_ep_recvmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
 
@@ -161,14 +162,15 @@ static ssize_t sock_ep_recvv(struct fid_ep *ep, const struct iovec *iov,
 		       void **desc, size_t count, fi_addr_t src_addr,
 		       void *context)
 {
-	struct fi_msg msg;
-	memset(&msg, 0, sizeof(msg));
-	msg.msg_iov = iov;
-	msg.desc = desc;
-	msg.iov_count = count;
-	msg.addr = src_addr;
-	msg.context = context;
-	msg.data = 0;
+	struct fi_msg msg = {
+		.msg_iov = iov,
+		.desc = desc,
+		.iov_count = count,
+		.addr = src_addr,
+		.context = context,
+		.data = 0,
+	};
+
 	return sock_ep_recvmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
 
@@ -290,16 +292,18 @@ err:
 static ssize_t sock_ep_send(struct fid_ep *ep, const void *buf, size_t len,
 		      void *desc, fi_addr_t dest_addr, void *context)
 {
-	struct fi_msg msg;
-	struct iovec msg_iov;
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-	msg.desc = &desc;
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.context = context;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg msg = {
+		.msg_iov = &msg_iov,
+		.desc = &desc,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.context = context,
+		.data = 0,
+	};
 
 	return sock_ep_sendmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
@@ -308,13 +312,15 @@ static ssize_t sock_ep_sendv(struct fid_ep *ep, const struct iovec *iov,
 		       void **desc, size_t count, fi_addr_t dest_addr,
 		       void *context)
 {
-	struct fi_msg msg;
-	memset(&msg, 0, sizeof(msg));
-	msg.msg_iov = iov;
-	msg.desc = desc;
-	msg.iov_count = count;
-	msg.addr = dest_addr;
-	msg.context = context;
+	struct fi_msg msg = {
+		.msg_iov = iov,
+		.desc = desc,
+		.iov_count = count,
+		.addr = dest_addr,
+		.context = context,
+		.data = 0,
+	};
+
 	return sock_ep_sendmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
 
@@ -322,18 +328,18 @@ static ssize_t sock_ep_senddata(struct fid_ep *ep, const void *buf, size_t len,
 			  void *desc, uint64_t data, fi_addr_t dest_addr,
 			  void *context)
 {
-	struct fi_msg msg;
-	struct iovec msg_iov;
-
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-
-	msg.msg_iov = &msg_iov;
-	msg.desc = desc;
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.context = context;
-	msg.data = data;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg msg = {
+		.msg_iov = &msg_iov,
+		.desc = desc,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.context = context,
+		.data = data,
+	};
 
 	return sock_ep_sendmsg(ep, &msg, FI_REMOTE_CQ_DATA | SOCK_USE_OP_FLAGS);
 }
@@ -341,34 +347,38 @@ static ssize_t sock_ep_senddata(struct fid_ep *ep, const void *buf, size_t len,
 static ssize_t sock_ep_inject(struct fid_ep *ep, const void *buf, size_t len,
 			fi_addr_t dest_addr)
 {
-	struct fi_msg msg;
-	struct iovec msg_iov;
-
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg msg = {
+		.msg_iov = &msg_iov,
+		.desc = NULL,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.context = NULL,
+		.data = 0,
+	};
 
 	return sock_ep_sendmsg(ep, &msg, FI_INJECT |
 			       SOCK_NO_COMPLETION | SOCK_USE_OP_FLAGS);
 }
 
-static ssize_t	sock_ep_injectdata(struct fid_ep *ep, const void *buf,
+static ssize_t sock_ep_injectdata(struct fid_ep *ep, const void *buf,
 				size_t len, uint64_t data, fi_addr_t dest_addr)
 {
-	struct fi_msg msg;
-	struct iovec msg_iov;
-
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.data = data;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg msg = {
+		.msg_iov = &msg_iov,
+		.desc = NULL,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.context = NULL,
+		.data = data,
+	};
 
 	return sock_ep_sendmsg(ep, &msg, FI_REMOTE_CQ_DATA | FI_INJECT |
 			       SOCK_NO_COMPLETION | SOCK_USE_OP_FLAGS);
@@ -476,21 +486,21 @@ static ssize_t sock_ep_trecv(struct fid_ep *ep, void *buf, size_t len,
 			void *desc, fi_addr_t src_addr, uint64_t tag,
 			uint64_t ignore, void *context)
 {
-	struct fi_msg_tagged msg;
-	struct iovec msg_iov;
+	struct iovec msg_iov = {
+		.iov_base = buf,
+		.iov_len = len,
+	};
+	struct fi_msg_tagged msg = {
+		.msg_iov = &msg_iov,
+		.desc = &desc,
+		.iov_count = 1,
+		.addr = src_addr,
+		.context = context,
+		.tag = tag,
+		.ignore = ignore,
+		.data = 0,
+	};
 
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = buf;
-	msg_iov.iov_len = len;
-
-	msg.msg_iov = &msg_iov;
-	msg.desc = &desc;
-	msg.iov_count = 1;
-	msg.addr = src_addr;
-	msg.context = context;
-	msg.tag = tag;
-	msg.ignore = ignore;
-	msg.data = 0;
 	return sock_ep_trecvmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
 
@@ -498,17 +508,17 @@ static ssize_t sock_ep_trecvv(struct fid_ep *ep, const struct iovec *iov,
 			       void **desc, size_t count, fi_addr_t src_addr,
 			       uint64_t tag, uint64_t ignore, void *context)
 {
-	struct fi_msg_tagged msg;
+	struct fi_msg_tagged msg = {
+		.msg_iov = iov,
+		.desc = desc,
+		.iov_count = count,
+		.addr = src_addr,
+		.context = context,
+		.tag = tag,
+		.ignore = ignore,
+		.data = 0,
+	};
 
-	memset(&msg, 0, sizeof(msg));
-	msg.msg_iov = iov;
-	msg.desc = desc;
-	msg.iov_count = count;
-	msg.addr = src_addr;
-	msg.context = context;
-	msg.tag = tag;
-	msg.ignore = ignore;
-	msg.data = 0;
 	return sock_ep_trecvmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
 
@@ -627,18 +637,20 @@ static ssize_t sock_ep_tsend(struct fid_ep *ep, const void *buf, size_t len,
 			void *desc, fi_addr_t dest_addr, uint64_t tag,
 			void *context)
 {
-	struct fi_msg_tagged msg;
-	struct iovec msg_iov;
-
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-	msg.desc = &desc;
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.context = context;
-	msg.tag = tag;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg_tagged msg = {
+		.msg_iov = &msg_iov,
+		.desc = &desc,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.tag = tag,
+		.ignore = 0,
+		.context = context,
+		.data = 0,
+	};
 
 	return sock_ep_tsendmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
@@ -647,15 +659,17 @@ static ssize_t sock_ep_tsendv(struct fid_ep *ep, const struct iovec *iov,
 			       void **desc, size_t count, fi_addr_t dest_addr,
 			       uint64_t tag, void *context)
 {
-	struct fi_msg_tagged msg;
+	struct fi_msg_tagged msg = {
+		.msg_iov = iov,
+		.desc = desc,
+		.iov_count = count,
+		.addr = dest_addr,
+		.tag = tag,
+		.ignore = 0,
+		.context = context,
+		.data = 0,
+	};
 
-	memset(&msg, 0, sizeof(msg));
-	msg.msg_iov = iov;
-	msg.desc = desc;
-	msg.iov_count = count;
-	msg.addr = dest_addr;
-	msg.context = context;
-	msg.tag = tag;
 	return sock_ep_tsendmsg(ep, &msg, SOCK_USE_OP_FLAGS);
 }
 
@@ -663,19 +677,20 @@ static ssize_t sock_ep_tsenddata(struct fid_ep *ep, const void *buf, size_t len,
 				void *desc, uint64_t data, fi_addr_t dest_addr,
 				uint64_t tag, void *context)
 {
-	struct fi_msg_tagged msg;
-	struct iovec msg_iov;
-
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-	msg.desc = desc;
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.context = context;
-	msg.data = data;
-	msg.tag = tag;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg_tagged msg = {
+		.msg_iov = &msg_iov,
+		.desc = desc,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.tag = tag,
+		.ignore = 0,
+		.context = context,
+		.data = data,
+	};
 
 	return sock_ep_tsendmsg(ep, &msg, FI_REMOTE_CQ_DATA | SOCK_USE_OP_FLAGS);
 }
@@ -683,16 +698,21 @@ static ssize_t sock_ep_tsenddata(struct fid_ep *ep, const void *buf, size_t len,
 static ssize_t sock_ep_tinject(struct fid_ep *ep, const void *buf, size_t len,
 				fi_addr_t dest_addr, uint64_t tag)
 {
-	struct fi_msg_tagged msg;
-	struct iovec msg_iov;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg_tagged msg = {
+		.msg_iov = &msg_iov,
+		.desc = NULL,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.tag = tag,
+		.ignore = 0,
+		.context = NULL,
+		.data = 0,
+	};
 
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.tag = tag;
 	return sock_ep_tsendmsg(ep, &msg, FI_INJECT |
 				SOCK_NO_COMPLETION | SOCK_USE_OP_FLAGS);
 }
@@ -701,18 +721,20 @@ static ssize_t	sock_ep_tinjectdata(struct fid_ep *ep, const void *buf,
 				size_t len, uint64_t data, fi_addr_t dest_addr,
 				uint64_t tag)
 {
-	struct fi_msg_tagged msg;
-	struct iovec msg_iov;
-
-	memset(&msg, 0, sizeof(msg));
-	msg_iov.iov_base = (void *) buf;
-	msg_iov.iov_len = len;
-	msg.msg_iov = &msg_iov;
-
-	msg.iov_count = 1;
-	msg.addr = dest_addr;
-	msg.data = data;
-	msg.tag = tag;
+	struct iovec msg_iov = {
+		.iov_base = (void *)buf,
+		.iov_len = len,
+	};
+	struct fi_msg_tagged msg = {
+		.msg_iov = &msg_iov,
+		.desc = NULL,
+		.iov_count = 1,
+		.addr = dest_addr,
+		.tag = tag,
+		.ignore = 0,
+		.context = NULL,
+		.data = data,
+	};
 
 	return sock_ep_tsendmsg(ep, &msg, FI_REMOTE_CQ_DATA | FI_INJECT |
 				SOCK_NO_COMPLETION | SOCK_USE_OP_FLAGS);
