@@ -34,7 +34,8 @@
 #define _VERBS_RDM_H
 
 #include <rdma/rdma_cma.h>
-#include "../uthash.h"
+#include <rdma/fi_atomic.h>
+#include <uthash.h>
 
 #include "../fi_verbs.h"
 #include "verbs_utils.h"
@@ -58,46 +59,39 @@
  *                                    IBV_RDM_ST_PKTYPE_MASK
  */
 
-// WR - work request
+/* WR - work request */
 #define FI_IBV_RDM_SERVICE_WR_MASK              ((uint64_t)0x1)
-#define FI_IBV_RDM_CHECK_SERVICE_WR_FLAG(value) \
+#define FI_IBV_RDM_CHECK_SERVICE_WR_FLAG(value)	\
         (value & FI_IBV_RDM_SERVICE_WR_MASK)
 
-#define FI_IBV_RDM_PACK_WR(value)               ((uint64_t)value)
-#define FI_IBV_RDM_UNPACK_WR(value)             ((void*)(uintptr_t)value)
+#define FI_IBV_RDM_PACK_WR(value)               ((uint64_t)(uintptr_t)(void *)value)
+#define FI_IBV_RDM_UNPACK_WR(value)             ((void *)(uintptr_t)value)
 
-#define FI_IBV_RDM_PACK_SERVICE_WR(value)       \
-        (((uint64_t)(uintptr_t)(void*)value) | FI_IBV_RDM_SERVICE_WR_MASK)
+#define FI_IBV_RDM_PACK_SERVICE_WR(value)					\
+        (((uint64_t)(uintptr_t)(void *)value) | FI_IBV_RDM_SERVICE_WR_MASK)
 
-#define FI_IBV_RDM_UNPACK_SERVICE_WR(value)     \
-        ((void*)(uintptr_t)(value & (~(FI_IBV_RDM_SERVICE_WR_MASK))))
+#define FI_IBV_RDM_UNPACK_SERVICE_WR(value)				\
+        ((void *)(uintptr_t)(value & (~(FI_IBV_RDM_SERVICE_WR_MASK))))
 
-// Send/Recv counters control
-
-#define FI_IBV_RDM_INC_SIG_POST_COUNTERS(_connection, _ep, _send_flags)	\
-do {                                                                	\
-	(_connection)->sends_outgoing++;                                \
-	(_ep)->posted_sends++;                                          \
-	(_send_flags) |= IBV_SEND_SIGNALED;                             \
-									\
-	VERBS_DBG(FI_LOG_CQ, "SEND_COUNTER++, conn %p, sends_outgoing %d\n",    \
-		_connection,                                            \
-		(_connection)->sends_outgoing);                         \
+/* Send/Recv counters control */
+#define FI_IBV_RDM_INC_SIG_POST_COUNTERS(_connection, _ep)				\
+do {                                                                			\
+	(_connection)->av_entry->sends_outgoing++;					\
+	(_ep)->posted_sends++;								\
+	VERBS_DBG(FI_LOG_CQ, "SEND_COUNTER++, conn %p, sends_outgoing %"PRIu32"\n",	\
+		  _connection, (_connection)->av_entry->sends_outgoing);		\
 } while (0)
 
-#define FI_IBV_RDM_DEC_SIG_POST_COUNTERS(_connection, _ep)		\
-do {									\
-	(_connection)->sends_outgoing--;				\
-	(_ep)->posted_sends--;						\
-									\
-	VERBS_DBG(FI_LOG_CQ, "SEND_COUNTER--, conn %p, sends_outgoing %d\n",	\
-			_connection, (_connection)->sends_outgoing);	\
-	assert((_ep)->posted_sends >= 0);				\
-	assert((_connection)->sends_outgoing >= 0);			\
+#define FI_IBV_RDM_DEC_SIG_POST_COUNTERS(_connection, _ep)				\
+do {											\
+	(_connection)->av_entry->sends_outgoing--;					\
+	(_ep)->posted_sends--;								\
+	VERBS_DBG(FI_LOG_CQ, "SEND_COUNTER--, conn %p, sends_outgoing %"PRIu32"\n",	\
+		  _connection, (_connection)->av_entry->sends_outgoing);		\
 } while (0)
 
-#define OUTGOING_POST_LIMIT(_connection, _ep)				\
-	((_connection)->sends_outgoing >= (_ep)->sq_wr_depth - 1)
+#define OUTGOING_POST_LIMIT(_connection, _ep)					\
+	((_connection)->av_entry->sends_outgoing >= (_ep)->sq_wr_depth - 1)
 
 #define PEND_POST_LIMIT(_ep)						\
 	((_ep)->posted_sends > 0.5 * (_ep)->scq_depth)
@@ -130,17 +124,19 @@ struct fi_ibv_rdm_rndv_header {
 	uint64_t src_addr;
 	uint64_t id; /* pointer to request on sender side */
 	uint64_t total_len;
-	uint32_t mem_key;
+	uint64_t mem_rkey;
 	uint32_t is_tagged;
 };
 
 struct fi_ibv_rdm_multi_request {
 	/* working request, will be renewed for every data arriving */
-	struct fi_ibv_rdm_request *prepost;
-	uint8_t *buf;
-	uint64_t len;
-	uint64_t offset;
-	uint64_t min_size;
+	struct dlist_entry		list_entry;
+	struct fi_ibv_rdm_request	*prepost;
+	struct fi_ibv_rdm_ep		*ep;
+	uint8_t				*buf;
+	uint64_t			len;
+	uint64_t			offset;
+	uint64_t			min_size;
 };
 
 struct fi_ibv_rdm_request {
@@ -157,7 +153,8 @@ struct fi_ibv_rdm_request {
 		ssize_t err; /* filled in case of moving to errcq */
 	} state;
 
-	struct fi_ibv_rdm_minfo minfo;
+	struct fi_ibv_rdm_ep	*ep;
+	struct fi_ibv_rdm_minfo	minfo;
 
 	/* User data: buffers, lens, imm, context */
 
@@ -195,16 +192,16 @@ struct fi_ibv_rdm_request {
 			/* registered buffer on sender side */
 			void* remote_addr;
 			/* registered mr of local src_addr */
-			struct ibv_mr *mr;
-			uint32_t rkey;
+			struct fi_ibv_mem_desc md;
+			uint64_t mr_rkey;
 		} rndv;
 		
 		/* RMA info */
 		struct {
-			struct ibv_mr* mr;
+			struct fi_ibv_mem_desc md;
+			uint64_t mr_rkey;
+			uint64_t mr_lkey;
 			uint64_t remote_addr;
-			uint32_t rkey;
-			uint32_t lkey;
 			enum ibv_wr_opcode opcode;
 		} rma;
 	};
@@ -265,23 +262,38 @@ struct fi_ibv_rdm_cntr {
 
 struct fi_ibv_rdm_ep {
 	struct fid_ep ep_fid;
-	struct fi_ibv_domain*	domain;
+	struct fi_ibv_domain	*domain;
 	struct slist_entry	list_entry;
-	struct fi_ibv_rdm_cq*	fi_scq;
-	struct fi_ibv_rdm_cq*	fi_rcq;
+	struct fi_ibv_rdm_cq	*fi_scq;
+	struct fi_ibv_rdm_cq	*fi_rcq;
 
-	struct fi_ibv_rdm_cntr*	send_cntr;
-	struct fi_ibv_rdm_cntr*	recv_cntr;
-	struct fi_ibv_rdm_cntr*	read_cntr;
-	struct fi_ibv_rdm_cntr*	write_cntr;
+	struct fi_ibv_rdm_cntr	*send_cntr;
+	struct fi_ibv_rdm_cntr	*recv_cntr;
+	struct fi_ibv_rdm_cntr	*read_cntr;
+	struct fi_ibv_rdm_cntr	*write_cntr;
 
-	struct fi_info*	info;
+	struct fi_info	*info;
+
+	struct util_buf_pool	*fi_ibv_rdm_request_pool;
+	struct util_buf_pool	*fi_ibv_rdm_multi_request_pool;
+	struct util_buf_pool	*fi_ibv_rdm_postponed_pool;
+	/*
+	 * extra buffer size equal eager buffer size,
+	 * it is used for any intermediate needs like
+	 * unexpected recv, pack/unpack noncontig messages, etc
+	 */
+	struct util_buf_pool	*fi_ibv_rdm_extra_buffers_pool;
+
+	struct dlist_entry	fi_ibv_rdm_posted_queue;
+	struct dlist_entry	fi_ibv_rdm_postponed_queue;
+	struct dlist_entry	fi_ibv_rdm_unexp_queue;
+	struct dlist_entry	fi_ibv_rdm_multi_recv_list;
 
 	size_t			addrlen;
-	struct rdma_addrinfo*	rai;
+	struct rdma_addrinfo	*rai;
 	struct sockaddr_in	my_addr;
 
-	struct fi_ibv_av*	av;
+	struct fi_ibv_av	*av;
 	int		tx_selective_completion;
 	int 		rx_selective_completion;
 	size_t 		min_multi_recv_size;
@@ -293,23 +305,22 @@ struct fi_ibv_rdm_ep {
 	 * It must generate work completion in receive CQ
 	 */
 	enum ibv_wr_opcode	eopcode;
-	struct ibv_cq*		scq;
-	struct ibv_cq*		rcq;
+	struct ibv_cq		*scq;
+	struct ibv_cq		*rcq;
 
 	int	buff_len;
 	int	n_buffs;
 	int	rq_wr_depth;    // RQ depth
 	int	sq_wr_depth;    // SQ depth
-	int	posted_sends;
-	int	posted_recvs;
+	uint32_t	posted_sends;
+	uint32_t	posted_recvs;
 	int	num_active_conns;
 	int	max_inline_rc;
 	int	rndv_threshold;
 	int	rndv_seg_size;
-	int	use_odp;
+	size_t	iov_per_rndv_thr;
 	int	scq_depth;
 	int	rcq_depth;
-	int	cqread_bunch_size;
 
 	int	is_closing;
 	int	recv_preposted_threshold;
@@ -321,7 +332,6 @@ enum fi_rdm_cm_conn_state {
 	FI_VERBS_CONN_REJECTED,
 	FI_VERBS_CONN_ESTABLISHED,
 	FI_VERBS_CONN_LOCAL_DISCONNECT,
-	FI_VERBS_CONN_REMOTE_DISCONNECT,
 	FI_VERBS_CONN_CLOSED
 };
 
@@ -334,10 +344,12 @@ enum fi_rdm_cm_role {
 
 struct fi_ibv_rdm_av_entry {
 	/* association of conn and EPs */
-	struct fi_ibv_rdm_conn		*conn_hash;
-	struct sockaddr_in		addr;
-	struct slist_entry		removed_next;
-	UT_hash_handle			hh;
+	struct fi_ibv_rdm_conn	*conn_hash;
+	struct sockaddr_in	addr;
+	uint32_t		sends_outgoing;
+	uint32_t		recv_preposted;
+	struct slist_entry	removed_next;
+	UT_hash_handle		hh;
 };
 
 struct fi_ibv_rdm_conn {
@@ -367,10 +379,10 @@ struct fi_ibv_rdm_conn {
 	struct dlist_entry postponed_requests_head;
 	struct fi_ibv_rdm_postponed_entry *postponed_entry;
 
-	struct ibv_mr *s_mr;
-	struct ibv_mr *r_mr;
-	struct ibv_mr *ack_mr;
-	struct ibv_mr *rma_mr;
+	struct fi_ibv_mem_desc s_md;
+	struct fi_ibv_mem_desc r_md;
+	struct fi_ibv_mem_desc ack_md;
+	struct fi_ibv_mem_desc rma_md;
 
 	uint32_t remote_sbuf_rkey;
 	uint32_t remote_rbuf_rkey;
@@ -379,8 +391,6 @@ struct fi_ibv_rdm_conn {
 	char *remote_rbuf_mem_reg;
 	struct fi_ibv_rdm_buf *remote_sbuf_head;
 
-	int sends_outgoing;
-	int recv_preposted;
 	/* counter for eager buffer releasing */
 	uint16_t recv_completions;
 	/* counter to control OOO behaviour, works in pair with recv_completions */
@@ -400,7 +410,14 @@ struct fi_ibv_rdm_postponed_entry {
 	struct fi_ibv_rdm_conn *conn;
 };
 
-extern struct util_buf_pool* fi_ibv_rdm_request_pool;
+struct fi_conn_priv_params {
+	char addr[FI_IBV_RDM_DFLT_ADDRLEN];
+
+	uint32_t rbuf_rkey;
+	uint32_t sbuf_rkey;
+	char *rbuf_mem_reg;
+	char *sbuf_mem_reg;
+};
 
 static inline void
 fi_ibv_rdm_set_buffer_status(struct fi_ibv_rdm_buf *buff, uint16_t status)
@@ -434,11 +451,11 @@ fi_ibv_rdm_push_buff_pointer(char *area_start, size_t area_size,
 	char *buff = (char*)(*rdm_buff);
 	char *buff_tmp = buff + offset;
 
-	VERBS_DBG(FI_LOG_EP_DATA, "old_pointer: %p\n", *buff);
+	VERBS_DBG(FI_LOG_EP_DATA, "old_pointer: %p\n", buff);
 
 	buff = buff_tmp < (area_start + area_size) ? buff_tmp : area_start;
 	
-	VERBS_DBG(FI_LOG_EP_DATA, "new_pointer: %p\n", *buff);
+	VERBS_DBG(FI_LOG_EP_DATA, "new_pointer: %p\n", buff);
 
 	*rdm_buff = (struct fi_ibv_rdm_buf *)buff;
 }
@@ -541,10 +558,13 @@ ssize_t fi_ibv_rdm_cm_progress(struct fi_ibv_rdm_ep *ep);
 ssize_t
 fi_ibv_rdm_start_overall_disconnection(struct fi_ibv_rdm_av_entry *av_entry);
 ssize_t fi_ibv_rdm_start_disconnection(struct fi_ibv_rdm_conn *conn);
+int fi_ibv_av_entry_alloc(struct fi_ibv_domain *domain,
+			  struct fi_ibv_rdm_av_entry **av_entry,
+			  void *addr);
 ssize_t fi_ibv_rdm_conn_cleanup(struct fi_ibv_rdm_conn *conn);
 ssize_t fi_ibv_rdm_overall_conn_cleanup(struct fi_ibv_rdm_av_entry *av_entry);
 ssize_t fi_ibv_rdm_start_connection(struct fi_ibv_rdm_ep *ep,
-                                struct fi_ibv_rdm_conn *conn);
+				    struct fi_ibv_rdm_conn *conn);
 ssize_t fi_ibv_rdm_repost_receives(struct fi_ibv_rdm_conn *conn,
 				   struct fi_ibv_rdm_ep *ep,
 				   int num_to_post);
@@ -646,41 +666,25 @@ fi_ibv_rdm_rma_get_buf_head(struct fi_ibv_rdm_conn *conn,
 }
 
 static inline int
-fi_ibv_rdm_check_connection(struct fi_ibv_rdm_conn *conn,
-			    struct fi_ibv_rdm_ep *ep)
+fi_ibv_rdm_check_connection(struct fi_ibv_rdm_conn *conn)
 {
-	const int status = (conn->state == FI_VERBS_CONN_ESTABLISHED);
-	if (!status) {
-		pthread_mutex_lock(&ep->domain->rdm_cm->cm_lock);
-		if (conn->state == FI_VERBS_CONN_ALLOCATED) {
-			fi_ibv_rdm_start_connection(ep, conn);
-		}
-		pthread_mutex_unlock(&ep->domain->rdm_cm->cm_lock);
-	}
-
-	return status;
+	return (conn->state == FI_VERBS_CONN_ESTABLISHED);
 }
 
 static inline struct fi_ibv_rdm_buf *
-fi_ibv_rdm_prepare_send_resources(struct fi_ibv_rdm_conn *conn,
-				  struct fi_ibv_rdm_ep *ep)
+fi_ibv_rdm_prepare_send_resources(struct fi_ibv_rdm_conn *conn)
 {
-	if (fi_ibv_rdm_check_connection(conn, ep)) {
-		return (!TSEND_RESOURCES_IS_BUSY(conn, ep)) ?
-			fi_ibv_rdm_get_sbuf_head(conn, ep) : NULL;
-	}
-	return NULL;
+	return (fi_ibv_rdm_check_connection(conn) &&
+		!TSEND_RESOURCES_IS_BUSY(conn, conn->ep)) ?
+		fi_ibv_rdm_get_sbuf_head(conn, conn->ep) : NULL;
 }
 
 static inline void *
-fi_ibv_rdm_rma_prepare_resources(struct fi_ibv_rdm_conn *conn,
-				 struct fi_ibv_rdm_ep *ep)
+fi_ibv_rdm_rma_prepare_resources(struct fi_ibv_rdm_conn *conn)
 {
-	if (fi_ibv_rdm_check_connection(conn, ep)) {
-		return (!RMA_RESOURCES_IS_BUSY(conn, ep)) ?
-			fi_ibv_rdm_rma_get_buf_head(conn, ep) : NULL;
-	}
-	return NULL;
+	return (fi_ibv_rdm_check_connection(conn) &&
+		!TSEND_RESOURCES_IS_BUSY(conn, conn->ep)) ?
+		fi_ibv_rdm_rma_get_buf_head(conn, conn->ep) : NULL;
 }
 
 static inline int
@@ -702,7 +706,7 @@ fi_ibv_rdm_process_send_wc(struct fi_ibv_rdm_ep *ep,
 	} else {
 		FI_IBV_DBG_OPCODE(wc->opcode, "SEND");
 		struct fi_ibv_rdm_request *request =
-			(void *)FI_IBV_RDM_UNPACK_WR(wc->wr_id);
+			FI_IBV_RDM_UNPACK_WR(wc->wr_id);
 
 		struct fi_ibv_rdm_tagged_send_completed_data data =
 			{ .ep = ep };
@@ -719,21 +723,23 @@ fi_ibv_rdm_process_err_send_wc(struct fi_ibv_rdm_ep *ep,
 	if (wc->status != IBV_WC_SUCCESS) {
 		struct fi_ibv_rdm_conn *conn;
 		if (FI_IBV_RDM_CHECK_SERVICE_WR_FLAG(wc->wr_id)) {
-			conn = FI_IBV_RDM_UNPACK_SERVICE_WR(
-					wc->wr_id);
+			conn = FI_IBV_RDM_UNPACK_SERVICE_WR(wc->wr_id);
 		} else {
 			struct fi_ibv_rdm_request *req =
-					(void *)wc->wr_id;
+				FI_IBV_RDM_UNPACK_WR(wc->wr_id);
 			conn = req->minfo.conn;
 			FI_IBV_RDM_DBG_REQUEST("to_pool: ", req,
 					       FI_LOG_DEBUG);
-			util_buf_release(fi_ibv_rdm_request_pool, req);
+			util_buf_release(ep->fi_ibv_rdm_request_pool, req);
 		}
 		VERBS_INFO(FI_LOG_EP_DATA, "got ibv_wc.status = %d:%s, "
-			   "pend_send: %d, connection: %p\n",
+			   "pend_send: %"PRIu32", connection: %p, request = %p (%s)\n",
 			   wc->status,
 			   ibv_wc_status_str(wc->status),
-			   ep->posted_sends, conn);
+			   ep->posted_sends, conn,
+			   (void *)wc->wr_id,
+			   FI_IBV_RDM_CHECK_SERVICE_WR_FLAG(wc->wr_id) ?
+				"service" : "not service");
 	}
 }
 
