@@ -63,16 +63,14 @@ struct fi_ibv_msg_ep;
 #define FI_IBV_RDM_DFLT_ADDRLEN	(sizeof (struct sockaddr_in))
 
 #define FI_IBV_RDM_CM_THREAD_TIMEOUT (100)
-#define FI_IBV_RDM_MEM_ALIGNMENT (64)
-#define FI_IBV_RDM_BUF_ALIGNMENT (4096) /* TODO: Page or MTU size */
 
 #define FI_IBV_RDM_TAGGED_DFLT_BUFFER_NUM (8)
 #define FI_IBV_RDM_DFLT_CQREAD_BUNCH_SIZE (FI_IBV_RDM_TAGGED_DFLT_BUFFER_NUM)
 
 #define FI_IBV_RDM_DFLT_BUFFER_SIZE					\
-	(3 * FI_IBV_RDM_BUF_ALIGNMENT)
+	(3 * FI_IBV_BUF_ALIGNMENT)
 
-#define FI_IBV_RDM_DFLT_BUFFERED_SSIZE					\
+#define FI_IBV_RDM_DFLT_BUFFERED_SIZE					\
 	(FI_IBV_RDM_DFLT_BUFFER_SIZE -					\
 	 FI_IBV_RDM_BUFF_SERVICE_DATA_SIZE -				\
 	 sizeof(struct fi_ibv_rdm_header))
@@ -127,19 +125,20 @@ do {										\
 	const size_t max_str_len = 1024;					\
 	char str[max_str_len];							\
 	snprintf(str, max_str_len,						\
-		"%s request: %p, eager_state: %s, rndv_state: %s,"		\
-		" err_state: %ld, tag: 0x%lx, len: %lu, rest: %lu,"		\
-		"context: %p, connection: %p\n",				\
-		prefix,								\
-		request,							\
-		fi_ibv_rdm_req_eager_state_to_str(request->state.eager),	\
-		fi_ibv_rdm_req_rndv_state_to_str(request->state.rndv),		\
-		request->state.err,						\
-		request->minfo.tag,						\
-		request->len,							\
-		request->rest_len,						\
-		request->context,						\
-		request->minfo.conn);						\
+		 "%s request: %p, eager_state: %s, rndv_state: %s,"		\
+		 " err_state: %ld, tag: 0x%lx, len: %lu, rest: %lu,"		\
+		 "context: %p, connection: %p ep: %p\n",			\
+		 prefix,							\
+		 request,							\
+		 fi_ibv_rdm_req_eager_state_to_str(request->state.eager),	\
+		 fi_ibv_rdm_req_rndv_state_to_str(request->state.rndv),		\
+		 request->state.err,						\
+		 request->minfo.tag,						\
+		 request->len,							\
+		 request->rest_len,						\
+		 request->context,						\
+		 request->minfo.conn,						\
+		 request->ep);							\
 										\
 	switch (level)								\
 	{									\
